/* $Id: main.c,v 1.5 2000/07/28 19:23:41 torsten Exp $ */

/*
** This file is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2, or (at your option)
** any later version.

** This file is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.

** You should have received a copy of the GNU General Public License
** along with GNU Emacs; see the file COPYING.  If not, write to
** the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
** Boston, MA 02111-1307, USA.

** Created by Torsten Hilbrich <Torsten.Hilbrich@gmx.net>
** 
** This file includes codes from wmppp.c which was created by the
** following authtors:
** Martijn Pieterse, Antoine Nulle, Thomas Apel (ISDN)

** Unfortunatly, there is to less space in the epplet window, so I had
** to remove those wonderful TX and RX indicators.
*/

#include <epplet.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <linux/isdn.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <errno.h>
#include <net/ppp_defs.h>
#include <net/if_ppp.h>
#include <stdlib.h>
#include <stdarg.h>

#ifndef IMAGEBASE
#define IMAGEBASE "./"
#endif

struct eppletinfo;

typedef int (*deviceSpecificCallback)(struct eppletinfo *);
typedef double time_unit;

enum states {
  OFFLINE=0,
  STANDBY=1,
  ONLINE=2,
  GOING_OFFLINE=3
};

struct rgb {
  int red;
  int green;
  int blue;
};

struct eppletinfo {
  Epplet_gadget onOffButton;
  Epplet_gadget imageArea;
  Epplet_gadget t_speed;
  Epplet_gadget t_time;
  Epplet_gadget b_close;
  Epplet_gadget b_help;
  Epplet_gadget b_config;

  Window config_win;

  Epplet_gadget config_interface;
  Epplet_gadget config_start;
  Epplet_gadget config_stop;
  Epplet_gadget config_update;
  Epplet_gadget config_timeout;

  long width;
  long height;

  /* the current connection state */
  enum states state;

  /* seconds between updates of transfer information */
  time_unit statisticUpdates;  

  /* seconds between online checks while being online */
  time_unit onlineChecks;

  /* seconds between online checks while being offline */
  time_unit offlineChecks;

  /* seconds to wait for the connection to be established or closed */
  time_unit timeout;

  /* the images the onoff button have for the certain states */
  char *images[4];

  /* the interface this info belongs to */
  char *interface;

  /* the command for starting the connection */
  char *startcmd;

  /* the command for stopping the connection */
  char *stopcmd;
  
  /* background */
  struct rgb backgroundColor;

  /* in transfer */
  struct rgb inColor;

  /* out transfer */
  struct rgb outColor;

  /* starttime */
  time_t starttime;

  deviceSpecificCallback initialize;
  deviceSpecificCallback checkOnline;
  deviceSpecificCallback transferInfo;

  /* information about the current connection */
  struct transferinfo *infos;
};

struct transferinfo {
  long inTransfer;
  long outTransfer;

  long *inHistory;
  long *outHistory;

  /* device specific information, can be isdninfo or pppinfo */
  void *deviceInfo;
};

struct isdninfo {
  int isdnFD;
  double lastUpdate;
  long inHist;
  long outHist;
};

void *getMem(size_t size)
{
  void *result = calloc(size, 1);
  if(result == 0) {
    fprintf(stderr, "Memory allocation failed\n");
    exit(1); 
  }
  return result;
}

void error(char *format, ...)
{
  char buffer[1024];
  va_list ap;
  int n;

  va_start(ap, format);
  n = vsnprintf(buffer, sizeof(buffer), format, ap);
  if(n > -1 && n < sizeof(buffer)) {
    /* display the error message */
    Epplet_dialog_ok(buffer);
  } else {
    Epplet_dialog_ok("Error message to long");
  }
  Esync();
  exit(1);
}

void syscall_error(char *name)
{
  error("%s: %s", name, strerror(errno));
}

void setImage(struct eppletinfo *data)
{
  Epplet_change_button_image(data->onOffButton, 
			     data->images[data->state]);
}

void setSpeed(struct eppletinfo *info, long speed)
{
  if(speed != -1) {
    char buffer[100];
    sprintf(buffer, "%ld", speed);
    Epplet_change_label(info->t_speed, buffer);
  } else {
    Epplet_change_label(info->t_speed, "");
  }
}

void setTime(struct eppletinfo *info, time_t speed)
{
  char buffer[100];
  sprintf(buffer, "%02ld:%02ld:%02ld", speed / 3600, (speed % 3600) / 60,
	  speed % 60);
  Epplet_change_label(info->t_time, buffer);
}

void setGraphic(struct eppletinfo *info)
{
  Window win = Epplet_get_drawingarea_window(info->imageArea);
  RGB_buf buf = Epplet_make_rgb_buf(info->width, info->height);
  long x,y;
  long *in = info->infos->inHistory;
  long *out = info->infos->outHistory;
  unsigned char *rgb = Epplet_get_rgb_pointer(buf), *ptr;

  long pixels_per_byte = info->height;

  for(x = 0; x < info->width; x++)
    if(in[x] + out[x] > pixels_per_byte)
      pixels_per_byte = in[x] + out[x];

  pixels_per_byte /= info->height;

  for(x = 0; x < info->width; x++) {
    for(y = 0; y < info->height; y++) {
      long val = (info->height - y + 1) * pixels_per_byte;
      ptr = rgb + y * info->width * 3 + x * 3;
      /*      printf("Val: %ld In: %ld Out: %ld\n", val, in[x], out[x]);*/


      if(val < in[x]) {
	ptr[0] = info->outColor.red;
	ptr[1] = info->outColor.green;
	ptr[2] = info->outColor.blue;
      } else if(val < (in[x] + out[x])) {
	ptr[0] = info->inColor.red;
	ptr[1] = info->inColor.green;
	ptr[2] = info->inColor.blue;
      } else {
	ptr[0] = info->backgroundColor.red;
	ptr[1] = info->backgroundColor.green;
	ptr[2] = info->backgroundColor.blue;
      }
    }
  }

  Epplet_paste_buf(buf, win, 0, 0);
  Epplet_free_rgb_buf(buf);

}



/***************************************
 * ISDN specific part.
 ***************************************/

/* BUG:
  currently this function checks only if any isdn network interface
  is active.
*/
int check_online_isdn(struct eppletinfo *data)
{
#ifdef IIOCNETGPN
  struct isdninfo *info = data->infos->deviceInfo;
  isdn_net_ioctl_phone phone;
  int rc;
  if(info->isdnFD == -1) {
    info->isdnFD = open("/dev/isdninfo", O_RDONLY | O_NDELAY);
    if(info->isdnFD == -1) {
      syscall_error("read /dev/isdninfo");
    }
  }
  
  memset(&phone, 0, sizeof(phone));
  strncpy(phone.name, data->interface, sizeof(phone.name));
  rc = ioctl(info->isdnFD, IIOCNETGPN, &phone);
  if(rc == 0) {
    return 1;
  } else if(errno == ENOTCONN) {
    return 0;
  } else if(errno == EINVAL) {
    error("You need Kernel 2.2.12, you should get the source package and recompile it");
  }
  return 0; /* dummy */
#else
  char temp[256];
  int result = 0;
  FILE *fp = fopen("/dev/isdninfo", "r");

  /*  printf("Checking online\n");*/

  if(fp) {
    while(fgets(temp, sizeof(temp), fp)) {
      if(strcmp("usage", temp) < 0) {
	result = (strncmp(temp + 7, "131", 3) == 0);
	break;
      }
    }
    fclose(fp);
    return result;
  } else {
    error("unable to read /dev/isdninfo\n");
    return 0; /* dummy */
  }
#endif
}

int initialize_isdn(struct eppletinfo *data)
{
  struct isdninfo *info = (struct isdninfo *)(data->infos->deviceInfo);

  if(info->isdnFD != -1) {
    close(info->isdnFD);
    info->isdnFD = -1;
  }
  info->lastUpdate = -1;
  info->inHist = 0;
  info->outHist = 0;
  return 0;
}

int transfer_info_isdn(struct eppletinfo *data)
{
  typedef struct {
    long ibytes;
    long obytes;
  } Siobytes;

  struct timeval tv;

  static Siobytes iobytes[ISDN_MAX_CHANNELS];

  long in, out;

  double now, diff;

  struct isdninfo *info = (struct isdninfo *)(data->infos->deviceInfo);

  if(info->isdnFD == -1) {
    /*    printf("Opening isdninfo\n");*/
    info->isdnFD = open("/dev/isdninfo", O_RDONLY | O_NDELAY);
    if(info->isdnFD == -1) {
      syscall_error("read /dev/isdninfo");
    }
  }
  
  if (ioctl(info->isdnFD, IIOCGETCPS, &iobytes) == -1)
    syscall_error("IIOCGETCPS");

  out = iobytes[0].obytes;
  in = iobytes[0].ibytes;

  gettimeofday(&tv, NULL);
  now = tv.tv_sec + (double)tv.tv_usec / 1e6;
  /*  printf("Now: %ld:%ld\n", tv.tv_sec, tv.tv_usec);*/

  if(info->lastUpdate > 0) {
    diff = now - info->lastUpdate;

    data->infos->inTransfer = (long)((in - info->inHist) / diff);
    data->infos->outTransfer = (long)((out - info->outHist) / diff);
    
    /*    printf("Diff %f, In: %ld, Out: %ld, iTransfer: %ld, oTransfer: %ld, iHist: %ld, oHist: %ld\n",
	   diff, in, out, data->infos->inTransfer, data->infos->outTransfer,
	   info->inHist, info->outHist);*/
  } else {
    data->infos->inTransfer = 0;
    data->infos->outTransfer = 0;
  }

  info->inHist = in;
  info->outHist = out;
  info->lastUpdate = now;
  /*
    data->infos->inTransfer = 5000 + random() % 3000;
    data->infos->outTransfer = 1000 + random() % 3000;
  */

  return 0;
}

void set_config(struct eppletinfo *data)
{
  data->statisticUpdates = atof(Epplet_query_config_def("statUpdates", "2.0"));
  data->onlineChecks = atof(Epplet_query_config_def("onlineChecks", "0.2"));
  data->offlineChecks = atof(Epplet_query_config_def("offlineChecks", "1.0"));
  data->timeout = atof(Epplet_query_config_def("timeout", "15.0"));
  data->images[0] = Epplet_query_config_def("offlineImage", 
					    IMAGEBASE "/green_dark.png");
  data->images[1] = Epplet_query_config_def("standbyImage", 
					    IMAGEBASE "/orange_light.png");
  data->images[2] = Epplet_query_config_def("onlineImage",
					    IMAGEBASE "/green_light.png");
  data->images[3] = Epplet_query_config_def("goingOfflineImage",
					    IMAGEBASE "/orange_light.png");
  data->interface = Epplet_query_config_def("interface", "ippp0");

  data->startcmd = Epplet_query_config_def("startcmd", "super isdnon");
  data->stopcmd = Epplet_query_config_def("stopcmd", "super isdnoff");

  data->backgroundColor.red = 0;
  data->backgroundColor.green = 0;
  data->backgroundColor.blue = 0;

  data->inColor.red = 50;
  data->inColor.green = 232;
  data->inColor.blue = 232;

  data->outColor.red = 50;
  data->outColor.green = 127;
  data->outColor.blue = 127;

  data->state = OFFLINE;
  data->infos = getMem(sizeof(struct transferinfo)); 
  data->infos->inTransfer = 0;
  data->infos->outTransfer = 0;
  data->infos->deviceInfo = 0;
  data->infos->inHistory = getMem(sizeof(long) * (data->width + 1));
  data->infos->outHistory = getMem(sizeof(long) * (data->width + 1));
  
  if(strncmp(data->interface, "ippp", 4) == 0) {
    data->initialize = initialize_isdn;
    data->checkOnline = check_online_isdn;
    data->transferInfo = transfer_info_isdn; 
    data->infos->deviceInfo = getMem(sizeof(struct isdninfo));
    ((struct isdninfo *)data->infos->deviceInfo)->isdnFD = -1;
    ((struct isdninfo *)data->infos->deviceInfo)->inHist = 0;
    ((struct isdninfo *)data->infos->deviceInfo)->outHist = 0;
    ((struct isdninfo *)data->infos->deviceInfo)->lastUpdate = -1;
  } else {
    error("Network interface %s not supported\n", data->interface);
  }
}

void free_config(struct eppletinfo *data)
{
  free(data->infos->deviceInfo);
  free(data->infos->inHistory);
  free(data->infos->outHistory);
  free(data->infos);
}

/*************************************
 * Interface part
 *************************************/

/* names for timers */
char *onlineTimer="online";
char *standbyTimer="standbyTimer";
char *updateStatTimer="updateStatTimer";
char *updateTimeTimer="updateTimeTimer";

void close_cb(void *data)
{
  Epplet_remember();
  Epplet_cleanup();
  exit(0);
}

void config_cancel_cb(void *data)
{
  struct eppletinfo *info = (struct eppletinfo *)data;

  Epplet_window_destroy(info->config_win);
  info->config_win = 0;
}

void config_apply_cb(void *data)
{
  struct eppletinfo *info = (struct eppletinfo *)data;

  Epplet_modify_config("interface", 
		       Epplet_textbox_contents(info->config_interface));
  Epplet_modify_config("startcmd", 
		       Epplet_textbox_contents(info->config_start));
  Epplet_modify_config("stopcmd", 
		       Epplet_textbox_contents(info->config_stop));
  Epplet_modify_config("statUpdates", 
		       Epplet_textbox_contents(info->config_update));
  Epplet_modify_config("timeout", 
		       Epplet_textbox_contents(info->config_timeout));
  
  set_config(info);
}

void config_ok_cb(void *data)
{
  config_apply_cb(data);
  config_cancel_cb(data);
}

void config_cb(void *data)
{
  struct eppletinfo *info = (struct eppletinfo *)data;

  char buffer[100];

  if(info->config_win) return;

  info->config_win = Epplet_create_window_config(230, 130, 
						 "E-pppMon Configuration",
						 config_ok_cb, data,
						 config_apply_cb, data,
						 config_cancel_cb, data);

  Epplet_gadget_show(Epplet_create_label(4,8, "Interface", 2));
  info->config_interface = 
    Epplet_create_textbox(NULL, info->interface, 
			  120, 4, 100, 20, 2, NULL, NULL);
  Epplet_gadget_show(info->config_interface);

  Epplet_gadget_show(Epplet_create_label(4,28, "Start command", 2));
  info->config_start =
    Epplet_create_textbox(NULL, info->startcmd, 
			  120, 24, 100, 20, 2, NULL, NULL);
  Epplet_gadget_show(info->config_start);

  Epplet_gadget_show(Epplet_create_label(4,48, "Stop command", 2));
  info->config_stop = Epplet_create_textbox(NULL, info->stopcmd, 
					    120, 44, 100, 20, 2, NULL, NULL);
  Epplet_gadget_show(info->config_stop);

  sprintf(buffer, "%.2f", info->statisticUpdates);

  Epplet_gadget_show(Epplet_create_label(4,68, "Update rate (seconds)", 2));
  info->config_update = Epplet_create_textbox(NULL, buffer, 
					      120, 64, 100, 20, 2, NULL, NULL);
  Epplet_gadget_show(info->config_update);

  sprintf(buffer, "%.2f", info->timeout);

  Epplet_gadget_show(Epplet_create_label(4,88, "Timeout (seconds)", 2));
  info->config_timeout = 
    Epplet_create_textbox(NULL, buffer, 
			  120, 84, 100, 20, 2, NULL, NULL);
  Epplet_gadget_show(info->config_timeout);

  Epplet_window_show(info->config_win);
  Epplet_window_pop_context();

}

void update_cb(void *);
void time_cb(void *);

void gone_online(struct eppletinfo *info)
{
  /*  printf("Gone online\n");*/
  info->starttime = time(0);
  Epplet_remove_timer(standbyTimer);
  info->state = ONLINE;
  setImage(info);
  update_cb(info);
  time_cb(info);
}

void gone_offline(struct eppletinfo *info)
{
  /*  printf("Gone offline\n");*/
  Epplet_remove_timer(standbyTimer);
  info->state = OFFLINE;
  setImage(info);
  Epplet_remove_timer(updateStatTimer);
  Epplet_remove_timer(updateTimeTimer);
  setSpeed(info, -1);
}

void check_online_cb(void *data)
{
  struct eppletinfo *info = (struct eppletinfo*)data;

  /*  printf("Current state: %d\n", info->state);*/
  if(info->checkOnline(info)) {
    if(info->state == STANDBY || info->state == OFFLINE)
      gone_online(info);
  } else {
    if(info->state == ONLINE || info->state == GOING_OFFLINE)
      gone_offline(info);
  }
  Epplet_timer(check_online_cb, info, 
	       info->state != OFFLINE ? info->onlineChecks 
                                      : info->offlineChecks, 
	       onlineTimer);
}

void timeout_cb(void *data)
{
  /*  printf("Timed out\n");*/
  gone_offline((struct eppletinfo*)data);
}

void go_online(struct eppletinfo *info)
{
  /*  printf("Go online\n");*/
  info->initialize(info);
  info->state = STANDBY;
  setImage(info);
  Epplet_timer(timeout_cb, info, info->timeout, standbyTimer);

  if(fork() == 0) {
    char *args[] = { "/bin/sh", "-c", info->startcmd, 0 };
    execv(args[0], args);
    syscall_error("exec startcmd");
  }
  Epplet_remove_timer(onlineTimer);
  Epplet_timer(check_online_cb, info, info->onlineChecks, 
	       onlineTimer);
}

void go_offline(struct eppletinfo *info)
{
  /*  printf("Go offline\n");*/
  info->state = GOING_OFFLINE;
  setImage(info);
  Epplet_timer(timeout_cb, info, info->timeout, standbyTimer);

  if(fork() == 0) {
    char *args[] = { "/bin/sh", "-c", info->stopcmd, 0 };
    execv(args[0], args);
    syscall_error("exec stopcmd");
  }
  Epplet_remove_timer(onlineTimer);
  Epplet_timer(check_online_cb, info, info->onlineChecks, onlineTimer);
}

void update_cb(void *data)
{
  struct eppletinfo *info = (struct eppletinfo*)data;
  int i;
  /*  printf("Update cb\n");*/

  info->transferInfo(info);

  info->infos->inHistory[info->width] = info->infos->inTransfer;
  info->infos->outHistory[info->width] = info->infos->outTransfer;

  /* move the history one step to the left */
  for(i = 1; i <= info->width; i++) {
    info->infos->inHistory[i - 1] = info->infos->inHistory[i];
    info->infos->outHistory[i - 1] = info->infos->outHistory[i];
  }

  setGraphic(info);
  setSpeed(info, info->infos->inTransfer + info->infos->outTransfer);

  if(info->checkOnline(info))
    Epplet_timer(update_cb, info, info->statisticUpdates, updateStatTimer);
}

void time_cb(void *data)
{
  struct eppletinfo *info = (struct eppletinfo*)data;

  setTime(info, time(0) - info->starttime);
  if(info->checkOnline(info))
    Epplet_timer(time_cb, info, 0.1, updateTimeTimer);
}

void onoff_cb(void *data)
{
  struct eppletinfo *info = (struct eppletinfo*)data;

  if(! info->checkOnline(info))
    go_online(info);
  else
    go_offline(info);
}

void in_cb(void *data, Window win)
{
  struct eppletinfo *info = (struct eppletinfo*)data;

  Epplet_gadget_show(info->b_close);
  Epplet_gadget_show(info->b_config);
  Epplet_gadget_show(info->b_help);
}

void out_cb(void *data, Window win)
{
  struct eppletinfo *info = (struct eppletinfo*)data;

  Epplet_gadget_hide(info->b_close);
  Epplet_gadget_hide(info->b_help);
  Epplet_gadget_hide(info->b_config);
}

void help_cb(void *data)
{
  Epplet_show_about(APPNAME);
}

void expose_cb(void *data, Window win, int x, int y, int w, int h)
{
  struct eppletinfo *info = (struct eppletinfo*)data;
  Window drawWin = Epplet_get_drawingarea_window(info->imageArea);
  if(win == drawWin) {
    setGraphic(info);
  }
}

#define HIST_WIDTH 40
#define HIST_HEIGHT 18

int main(int argc, char *argv[])
{
  struct eppletinfo info;

  Epplet_gadget b_close, b_config, b_help, t_speed, t_time, b_onoff, g_draw;
    

  /* 
   * First set the priority of this process, ignoring failures
   * if the currenty priority is larger.
   */
  setpriority(PRIO_PROCESS, getpid(), 10);

  /* At exit cleanup everything */
  atexit(Epplet_cleanup);

  /* The main window */
  Epplet_Init(APPNAME, "0.1", "Enlightenment ippp and ppp monitor",
	      3, 3, argc, argv, 0);

  b_onoff = Epplet_create_image_button("", 34, 3, 10, 10, onoff_cb, &info);

  b_close = Epplet_create_std_button("CLOSE", 2, 2, close_cb, &info);

  b_config = Epplet_create_std_button("CONFIGURE", 2, 34, config_cb, &info);

  b_help = Epplet_create_std_button("HELP", 14, 34, help_cb, &info);

  t_speed = Epplet_create_label(2, 2, "", 2);
  t_time = Epplet_create_label(4, 35, "", 2);

  g_draw = Epplet_create_drawingarea(2, 13, HIST_WIDTH + 4, HIST_HEIGHT + 4);
  
  Epplet_gadget_show(t_speed);
  Epplet_gadget_show(t_time);
  Epplet_gadget_show(b_onoff);
  Epplet_gadget_show(g_draw);

  Epplet_register_focus_in_handler(in_cb, &info);
  Epplet_register_focus_out_handler(out_cb, &info);

  info.t_speed = t_speed;
  info.t_time = t_time;
  info.imageArea = g_draw;
  info.onOffButton = b_onoff;
  info.b_close = b_close;
  info.b_config = b_config;
  info.b_help = b_help;
  info.width = HIST_WIDTH;
  info.height = HIST_HEIGHT;

  info.config_win = 0;

  Epplet_load_config();
  set_config(&info);

  setImage(&info);

  /*  update_cb(&info);*/

  Epplet_timer(check_online_cb, &info, info.offlineChecks, onlineTimer);

  Epplet_register_expose_handler(expose_cb, &info);
  
  Epplet_show();
  Epplet_Loop();

  return 0;
}
